# -*- coding: utf-8 -*-
from playwright.async_api import Page as AsyncPage
from playwright.sync_api import Page as SyncPage
from playwright_stealth.core import StealthConfig
from playwright_stealth.properties import Properties, BrowserType


def combine_scripts(properties: Properties, config: StealthConfig):
    """Combines the scripts for the page based on the properties and config."""

    scripts = []

    for script in (config or StealthConfig()).enabled_scripts(properties):
        scripts.append(script)
    return "\n".join(scripts)


def generate_stealth_headers_sync(properties: Properties, page: SyncPage):
    """Generates the stealth headers for the page by replacing the original headers with the spoofed ones for every request."""
    page.set_extra_http_headers(properties.as_dict()["header"])


async def generate_stealth_headers_async(properties: Properties, page: AsyncPage):
    """Generates the stealth headers for the page by replacing the original headers with the spoofed ones for every request."""
    await page.set_extra_http_headers(properties.as_dict()["header"])


def stealth_sync(page: SyncPage, config: StealthConfig = None):
    """teaches synchronous playwright Page to be stealthy like a ninja!"""
    properties = Properties(browser_type=config.browser_type if config else BrowserType.CHROME)
    combined_script = combine_scripts(properties, config)
    generate_stealth_headers_sync(properties, page)

    page.add_init_script(combined_script)


async def stealth_async(page: AsyncPage, config: StealthConfig = None):
    """teaches asynchronous playwright Page to be stealthy like a ninja!"""
    properties = Properties(browser_type=config.browser_type if config else BrowserType.CHROME)
    combined_script = combine_scripts(properties, config)
    await generate_stealth_headers_async(properties, page)

    await page.add_init_script(combined_script)



from dataclasses import dataclass
from typing import List


@dataclass
class HeaderProperties:
    """Class for the header properties. We will take the Sec_Fetch_Site, Sec_Fetch_Mode and Sec_Fetch_Dest from the original headers."""

    # Headers passed by library
    user_agent: str
    accept_language: str
    accept_encoding: str
    accept: str
    referer: str

    # Self generated headers
    origin: str
    sec_ch_ua: str
    sec_ch_ua_mobile: str
    sec_ch_ua_platform: str
    sec_ch_ua_form_factors: str
    dnt: str

    def __init__(
        self,
        brands: List[dict],
        dnt: str,
        client_hint_headers_enabled: bool = True,
        **kwargs,
    ):
        # Passed by library
        self.user_agent = kwargs["User-Agent"]
        self.accept_language = kwargs["Accept-language"]
        self.accept_encoding = kwargs["Accept-encoding"]
        self.accept = kwargs["Accept"]
        self.referer = kwargs["Referer"]

        # # Shared properties
        self.dnt = dnt

        # # Self generated headers
        if client_hint_headers_enabled:
            self.sec_ch_ua = self._generate_sec_ch_ua(brands)
            self.sec_ch_ua_mobile = self._generate_sec_ch_ua_mobile()
            self.sec_ch_ua_platform = self._generate_sec_ch_ua_platform()
            self.sec_ch_ua_form_factors = self._generate_sec_ch_ua_form_factors()

    def _generate_sec_ch_ua_platform(self) -> str:
        """Generates the Sec_Ch_Ua_Platform based on the user agent platform."""

        is_mac = "Macintosh" in self.user_agent
        is_windows = "Windows" in self.user_agent
        is_linux = "Linux" in self.user_agent

        if is_mac:
            return "macOS"
        elif is_windows:
            return "Windows"
        elif is_linux:
            return "Linux"
        else:
            return "Unknown"

    def _generate_sec_ch_ua(self, brands: List[dict]) -> str:
        """Generates the Sec_Ch_Ua based brands generated"""
        merged_brands = "".join([f'"{brand["brand"]}";v="{brand["version"]}",' for brand in brands])
        return merged_brands

    def _generate_sec_ch_ua_form_factors(self) -> str:
        """Generates the Sec_Ch_Ua_Form_Factors based on the user agent."""

        return "desktop"

    def _generate_sec_ch_ua_mobile(self) -> str:
        """Generates the Sec_Ch_Ua_Mobile based on the user agent."""

        return "?0"

    def as_dict(self) -> dict:
        # Convert all keys to kebab case and return a new dictionary
        return {key.replace("_", "-").lower(): value for key, value in self.__dict__.items()}



import json
from dataclasses import dataclass
from typing import Dict, Tuple, Optional
import os
from playwright_stealth.properties import Properties, BrowserType


def from_file(name) -> str:
    """Read script from ./js directory"""
    filename = os.path.join(os.path.dirname(os.path.dirname(__file__)), "js", name)
    with open(filename, encoding="utf-8") as f:
        return f.read()


SCRIPTS: Dict[str, str] = {
    "chrome_csi": from_file("chrome.csi.js"),
    "chrome_app": from_file("chrome.app.js"),
    "chrome_runtime": from_file("chrome.runtime.js"),
    "chrome_load_times": from_file("chrome.load.times.js"),
    "generate_magic_arrays": from_file("generate.magic.arrays.js"),
    "iframe_content_window": from_file("iframe.contentWindow.js"),
    "media_codecs": from_file("media.codecs.js"),
    "navigator_vendor": from_file("navigator.vendor.js"),
    "navigator_plugins": from_file("navigator.plugins.js"),
    "navigator_permissions": from_file("navigator.permissions.js"),
    "navigator_languages": from_file("navigator.languages.js"),
    "navigator_user_agent": from_file("navigator.userAgent.js"),
    "navigator_hardware_concurrency": from_file("navigator.hardwareConcurrency.js"),
    "outerdimensions": from_file("window.outerdimensions.js"),
    "utils": from_file("utils.js"),
    "webdriver": from_file("navigator.webdriver.js"),
    "webgl_vendor": from_file("webgl.vendor.js"),
}


@dataclass
class StealthConfig:
    """
    Playwright stealth configuration that applies stealth strategies to playwright page objects.
    The stealth strategies are contained in ./js package and are basic javascript scripts that are executed
    on every page.goto() called.
    Note:
        All init scripts are combined by playwright into one script and then executed this means
        the scripts should not have conflicting constants/variables etc. !
        This also means scripts can be extended by overriding enabled_scripts generator:
        ```
        @property
        def enabled_scripts():
            yield 'console.log("first script")'
            yield from super().enabled_scripts()
            yield 'console.log("last script")'
        ```
    """

    # load script options
    webdriver: bool = True
    webgl_vendor: bool = True
    chrome_app: bool = True
    chrome_csi: bool = True
    chrome_load_times: bool = True
    chrome_runtime: bool = True
    iframe_content_window: bool = True
    media_codecs: bool = True
    navigator_hardware_concurrency: int = 4
    navigator_languages: bool = True
    navigator_permissions: bool = True
    navigator_plugins: bool = True
    navigator_user_agent: bool = True
    navigator_vendor: bool = True
    outerdimensions: bool = True
    browser_type: BrowserType = BrowserType.CHROME

    # options
    vendor: str = "Intel Inc."
    renderer: str = "Intel Iris OpenGL Engine"
    nav_vendor: str = "Google Inc."
    nav_user_agent: str = None
    nav_platform: str = None
    languages: Tuple[str, str] = ("en-US", "en")
    run_on_insecure_origins: Optional[bool] = None

    def enabled_scripts(self, properties: Properties):
        """
        Generate the scripts to be executed.
        """

        opts = json.dumps(properties.as_dict())

        # defined options constant
        yield f"const opts = {opts}"
        # init utils and generate_magic_arrays helper
        yield SCRIPTS["utils"]
        yield SCRIPTS["generate_magic_arrays"]

        if self.chrome_app:
            yield SCRIPTS["chrome_app"]
        if self.chrome_csi:
            yield SCRIPTS["chrome_csi"]
        if self.chrome_load_times:
            yield SCRIPTS["chrome_load_times"]
        if self.chrome_runtime:
            yield SCRIPTS["chrome_runtime"]
        if self.iframe_content_window:
            yield SCRIPTS["iframe_content_window"]
        if self.media_codecs:
            yield SCRIPTS["media_codecs"]
        if self.navigator_languages:
            yield SCRIPTS["navigator_languages"]
        if self.navigator_permissions:
            yield SCRIPTS["navigator_permissions"]
        if self.navigator_plugins:
            yield SCRIPTS["navigator_plugins"]
        if self.navigator_user_agent:
            yield SCRIPTS["navigator_user_agent"]
        if self.navigator_vendor:
            yield SCRIPTS["navigator_vendor"]
        if self.webdriver:
            yield SCRIPTS["webdriver"]
        if self.outerdimensions:
            yield SCRIPTS["outerdimensions"]
        if self.webgl_vendor:
            yield SCRIPTS["webgl_vendor"]



import random
from dataclasses import dataclass


@dataclass
class WebGlProperties:
    vendor: str
    renderer: str

    webgl_properties = [
        {"vendor": "Intel Inc.", "renderer": "Intel Iris OpenGL Engine"},
        {"vendor": "AMD", "renderer": "AMD Radeon Pro 5600M OpenGL Engine"},
        {
            "vendor": "NVIDIA",
            "renderer": "NVIDIA GeForce GTX 1660 Ti OpenGL Engine",
        },
        {"vendor": "Apple Inc.", "renderer": "Apple M1 OpenGL Engine"},
        {"vendor": "Qualcomm Inc.", "renderer": "Qualcomm Adreno OpenGL Engine"},
    ]

    def __init__(self):
        webgl_prop = self._generate_webgl_prop()
        self.vendor = webgl_prop["vendor"]
        self.renderer = webgl_prop["renderer"]

    def _generate_webgl_prop(self):
        """Generates a WebGL property containing both vendor and renderer."""
        return random.choice(self.webgl_properties)

    def as_dict(self):
        return self.__dict__



from dataclasses import dataclass
from typing import Tuple
import random


@dataclass
class ViewportProperties:
    width: int
    height: int
    outerWidth: int
    outerHeight: int
    innerWidth: int
    innerHeight: int

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

        self.width, self.height = self._generate_viewport_dimensions()
        self.outerWidth, self.outerHeight = self._generate_outer_dimensions()
        self.innerWidth, self.innerHeight = self._generate_inner_dimensions()

    def _generate_viewport_dimensions(self) -> Tuple[int, int]:
        return 1920 + random.randint(-100, 100), 1080 + random.randint(-100, 100)

    def _generate_outer_dimensions(self) -> Tuple[int, int]:
        return self.width, self.height

    def _generate_inner_dimensions(self) -> Tuple[int, int]:
        return (
            self.width - random.randint(0, 20),
            self.height - random.randint(0, 20),
        )

    def as_dict(self) -> dict:
        return self.__dict__



from dataclasses import dataclass
from typing import List


@dataclass
class NavigatorProperties:
    """Class for the navigator properties."""

    userAgent: str
    platform: str
    language: str
    languages: List[str]
    appVersion: str
    vendor: str
    deviceMemory: int
    hardwareConcurrency: int
    maxTouchPoints: int
    doNotTrack: str
    brands: List[dict]
    mobile: bool

    def __init__(self, brands: List[dict], dnt: str, **kwargs):
        self.userAgent = kwargs["User-Agent"]

        # Shared properties
        self.brands = brands
        self.doNotTrack = dnt

        # Generate properties
        self.platform = self._generate_platform(kwargs["User-Agent"])
        self.language = self._generate_language()
        self.languages = self._generate_languages(kwargs["Accept-language"])
        self.appVersion = self._generate_app_version(kwargs["User-Agent"])
        self.vendor = self._generate_vendor(kwargs["User-Agent"])
        self.deviceMemory = self._generate_device_memory(kwargs["User-Agent"])
        self.hardwareConcurrency = self._generate_hardware_concurrency(self.deviceMemory)
        self.maxTouchPoints = self._generate_max_touch_points()
        self.mobile = self._generate_mobile()
        self.productSub = self._generate_product_sub(kwargs["User-Agent"])

    def _generate_platform(self, user_agent: str) -> str:
        """Generates the platform based on the user agent."""

        # Change regarding this article: \
        # https://filipvitas.medium.com/how-to-set-user-agent-header-with-puppeteer-js-and-not-fail-28c7a02165da
        if "Macintosh" in user_agent:
            return "MacIntel"
        elif "Linux" in user_agent:
            return "Linux x86_x64"
        else:
            return "Win64"

    def _generate_language(self) -> str:
        """Generates the language based on the accept language."""

        return "en-US"

    def _generate_languages(self, accept_language: str) -> List[str]:
        """Generates the languages based on the accept language."""

        languages_with_quality = accept_language.split(",")
        languages = [language.split(";")[0] for language in languages_with_quality]
        return languages

    def _generate_app_version(self, user_agent: str) -> str:
        """Generates the app version based on the user agent."""

        version_part = user_agent.split("/", 1)[1]
        return version_part

    def _generate_vendor(self, user_agent: str) -> str:
        """Generates the vendor based on the user agent."""

        if "Chrome" in user_agent:
            return "Google Inc."
        elif "Firefox" in user_agent:
            return ""

        return "Google Inc."

    def _generate_device_memory(self, user_agent: str) -> int:
        """Generates the device memory."""

        return None if "Firefox" in user_agent else 8

    def _generate_hardware_concurrency(self, device_memory: int) -> int:
        """Generates the hardware concurrency."""

        return device_memory

    def _generate_max_touch_points(self) -> int:
        """Generates the max touch points. Default is 0 since this is a desktop browser."""

        return 0

    def _generate_mobile(self) -> bool:
        """Generates the mobile flag."""

        return False

    def _generate_product_sub(self, user_agent: str) -> int:
        """Generate product sub depending on the Browser"""
        return 20100101 if "Firefox" in user_agent else 20030107

    def as_dict(self) -> dict:
        return self.__dict__



import re
import random
from dataclasses import dataclass
from fake_http_header import FakeHttpHeader
from ._header_properties import HeaderProperties
from ._navigator_properties import NavigatorProperties
from ._viewport_properties import ViewportProperties
from ._webgl_properties import WebGlProperties
from enum import Enum


class BrowserType(Enum):
    CHROME = "chrome"
    FIREFOX = "firefox"
    SAFARI = "safari"


@dataclass
class Properties:

    header: HeaderProperties
    navigator: NavigatorProperties
    viewport: ViewportProperties
    webgl: WebGlProperties
    runOnInsecureOrigins: bool

    def __init__(self, browser_type: BrowserType = BrowserType.CHROME):
        spoofed_headers = FakeHttpHeader(domain_code="com", browser=browser_type.value)

        # Generate shared properties
        brands = self._generate_brands(spoofed_headers.user_agent, browser=browser_type.value)
        dnt = self._generate_dnt()

        # Generate properties
        self.header = HeaderProperties(
            brands=brands,
            dnt=dnt,
            client_hint_headers_enabled=browser_type
            is not BrowserType.FIREFOX,  # Firefox does not support client hints
            **spoofed_headers.as_header_dict(),
        )
        self.navigator = NavigatorProperties(
            brands=brands, dnt=dnt, **spoofed_headers.as_header_dict()
        )
        self.viewport = ViewportProperties()
        self.webgl = WebGlProperties()
        self.runOnInsecureOrigins = None

    def _generate_brands(self, user_agent: str, browser: str = "chrome") -> str:
        """Generates the brands based on the referer."""

        configs = {
            "chrome": {
                "regex": r"Chrome/(\d+)",
                "brands": ["Chromium", "Google Chrome"],
            },
            "firefox": {
                "regex": r"Firefox/(\d+)",
                "brands": ["Firefox", "Firefox"],
            },
            "safari": {
                "regex": r"Safari/(\d+)",
                "brands": ["Safari", "Apple WebKit"],
            },
        }

        config = configs[browser]
        pattern = config["regex"]

        browser_with_version = re.search(pattern, user_agent)
        version = browser_with_version.group(1)
        seed = int(version.split(".")[0])

        order = [
            [0, 1, 2],
            [0, 2, 1],
            [1, 0, 2],
            [1, 2, 0],
            [2, 0, 1],
            [2, 1, 0],
        ][seed % 6]

        escaped_chars = [" ", " ", ";"]

        greasey_brand = (
            f"{escaped_chars[order[0]]}Not{escaped_chars[order[1]]}A{escaped_chars[order[2]]}Brand"
        )

        greased_brand_version_list = [{}, {}, {}]

        greased_brand_version_list[order[0]] = {
            "brand": greasey_brand,
            "version": "99",
        }

        greased_brand_version_list[order[1]] = {
            "brand": config["brands"][0],
            "version": seed,
        }

        greased_brand_version_list[order[2]] = {
            "brand": config["brands"][1],
            "version": seed,
        }

        return greased_brand_version_list

    def _generate_dnt(self) -> str:
        """Randomly generates a 0 or 1."""

        return str(random.randint(0, 1))

    def as_dict(self) -> dict:
        """Returns the properties as a dictionary."""

        return {
            "header": self.header.as_dict(),
            "viewport": self.viewport.as_dict(),
            "navigator": self.navigator.as_dict(),
            "webgl": self.webgl.as_dict(),
        }
